# This file was auto-generated by Fern from our API Definition.

import typing
from json.decoder import JSONDecodeError

from ..core.api_error import ApiError
from ..core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ..core.http_response import AsyncHttpResponse, HttpResponse
from ..core.jsonable_encoder import jsonable_encoder
from ..core.pydantic_utilities import parse_obj_as
from ..core.request_options import RequestOptions
from ..errors.unprocessable_entity_error import UnprocessableEntityError
from ..types.branch_response import BranchResponse
from ..types.http_validation_error import HttpValidationError
from ..types.redeploy_response import RedeployResponse
from ..types.rollback_response import RollbackResponse

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class RawBranchesClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._client_wrapper = client_wrapper

    def agents_environments_rollback_create(
        self,
        namespace_slug: str,
        agent_name: str,
        env_name: str,
        *,
        target_version_id: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[RollbackResponse]:
        """
        Roll back an environment to a previous version.

            Resolves environment to branch via branch rules.
            Only works for environments with a single, literal branch rule (no wildcards).

            Key behavior:
            - EnvVar table is NOT modified (remains pending state for next deploy)
            - Uses Version.secrets_snapshot as deployed state

        Parameters
        ----------
        namespace_slug : str

        agent_name : str

        env_name : str

        target_version_id : typing.Optional[str]
            Version ID to rollback to (defaults to previous version if not specified)

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[RollbackResponse]
            Successful Response
        """
        _response = self._client_wrapper.httpx_client.request(
            f"agents/{jsonable_encoder(namespace_slug)}/{jsonable_encoder(agent_name)}/environments/{jsonable_encoder(env_name)}/rollback",
            method="POST",
            json={
                "target_version_id": target_version_id,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    RollbackResponse,
                    parse_obj_as(
                        type_=RollbackResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        HttpValidationError,
                        parse_obj_as(
                            type_=HttpValidationError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def retrieve(
        self, branch_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[BranchResponse]:
        """
        Get branch details by ID.

            Used by CLI to poll for branch status after calling POST /agents/deploy.
            Poll until status is READY (success) or FAILED/TIMEOUT (failure).

        Parameters
        ----------
        branch_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[BranchResponse]
            Successful Response
        """
        _response = self._client_wrapper.httpx_client.request(
            f"branches/{jsonable_encoder(branch_id)}",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    BranchResponse,
                    parse_obj_as(
                        type_=BranchResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        HttpValidationError,
                        parse_obj_as(
                            type_=HttpValidationError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def redeploy(
        self, branch_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[RedeployResponse]:
        """
        Redeploy a branch with the current secrets from the EnvVar table.

            Creates a new Version with:
            - Same image as current version
            - Fresh secrets_snapshot from current EnvVars

            Use this after updating secrets to apply them to a running branch.

        Parameters
        ----------
        branch_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[RedeployResponse]
            Successful Response
        """
        _response = self._client_wrapper.httpx_client.request(
            f"branches/{jsonable_encoder(branch_id)}/redeploy",
            method="POST",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    RedeployResponse,
                    parse_obj_as(
                        type_=RedeployResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        HttpValidationError,
                        parse_obj_as(
                            type_=HttpValidationError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def rollback(
        self,
        branch_id: str,
        *,
        target_version_id: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[RollbackResponse]:
        """
        Roll back a branch to a previous version by branch ID.

            Key behavior:
            - EnvVar table is NOT modified (remains pending state)
            - Uses Version.secrets_snapshot as deployed state

        Parameters
        ----------
        branch_id : str

        target_version_id : typing.Optional[str]
            Version ID to rollback to (defaults to previous version if not specified)

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[RollbackResponse]
            Successful Response
        """
        _response = self._client_wrapper.httpx_client.request(
            f"branches/{jsonable_encoder(branch_id)}/rollback",
            method="POST",
            json={
                "target_version_id": target_version_id,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    RollbackResponse,
                    parse_obj_as(
                        type_=RollbackResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        HttpValidationError,
                        parse_obj_as(
                            type_=HttpValidationError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)


class AsyncRawBranchesClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._client_wrapper = client_wrapper

    async def agents_environments_rollback_create(
        self,
        namespace_slug: str,
        agent_name: str,
        env_name: str,
        *,
        target_version_id: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[RollbackResponse]:
        """
        Roll back an environment to a previous version.

            Resolves environment to branch via branch rules.
            Only works for environments with a single, literal branch rule (no wildcards).

            Key behavior:
            - EnvVar table is NOT modified (remains pending state for next deploy)
            - Uses Version.secrets_snapshot as deployed state

        Parameters
        ----------
        namespace_slug : str

        agent_name : str

        env_name : str

        target_version_id : typing.Optional[str]
            Version ID to rollback to (defaults to previous version if not specified)

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[RollbackResponse]
            Successful Response
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"agents/{jsonable_encoder(namespace_slug)}/{jsonable_encoder(agent_name)}/environments/{jsonable_encoder(env_name)}/rollback",
            method="POST",
            json={
                "target_version_id": target_version_id,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    RollbackResponse,
                    parse_obj_as(
                        type_=RollbackResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        HttpValidationError,
                        parse_obj_as(
                            type_=HttpValidationError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def retrieve(
        self, branch_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[BranchResponse]:
        """
        Get branch details by ID.

            Used by CLI to poll for branch status after calling POST /agents/deploy.
            Poll until status is READY (success) or FAILED/TIMEOUT (failure).

        Parameters
        ----------
        branch_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[BranchResponse]
            Successful Response
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"branches/{jsonable_encoder(branch_id)}",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    BranchResponse,
                    parse_obj_as(
                        type_=BranchResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        HttpValidationError,
                        parse_obj_as(
                            type_=HttpValidationError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def redeploy(
        self, branch_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[RedeployResponse]:
        """
        Redeploy a branch with the current secrets from the EnvVar table.

            Creates a new Version with:
            - Same image as current version
            - Fresh secrets_snapshot from current EnvVars

            Use this after updating secrets to apply them to a running branch.

        Parameters
        ----------
        branch_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[RedeployResponse]
            Successful Response
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"branches/{jsonable_encoder(branch_id)}/redeploy",
            method="POST",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    RedeployResponse,
                    parse_obj_as(
                        type_=RedeployResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        HttpValidationError,
                        parse_obj_as(
                            type_=HttpValidationError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def rollback(
        self,
        branch_id: str,
        *,
        target_version_id: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[RollbackResponse]:
        """
        Roll back a branch to a previous version by branch ID.

            Key behavior:
            - EnvVar table is NOT modified (remains pending state)
            - Uses Version.secrets_snapshot as deployed state

        Parameters
        ----------
        branch_id : str

        target_version_id : typing.Optional[str]
            Version ID to rollback to (defaults to previous version if not specified)

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[RollbackResponse]
            Successful Response
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"branches/{jsonable_encoder(branch_id)}/rollback",
            method="POST",
            json={
                "target_version_id": target_version_id,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    RollbackResponse,
                    parse_obj_as(
                        type_=RollbackResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        HttpValidationError,
                        parse_obj_as(
                            type_=HttpValidationError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)



"""SDK version detection utilities."""

from __future__ import annotations

import importlib.metadata
from functools import lru_cache

MIN_CODEXADAPTER_OPENAI_AGENTS_VERSION = "0.9.0"


def get_claude_sdk_version() -> str:
    """Get the Claude SDK version, with fallback."""
    try:
        import claude_agent_sdk

        return claude_agent_sdk.__version__
    except (ImportError, AttributeError):
        return "0.0.0-unknown"


def get_codex_sdk_version() -> str:
    """Get canonical Codex adapter version from installed openai-agents package."""
    detected_version = _detect_openai_agents_version()
    if detected_version:
        return detected_version

    # Fallback to the minimum 0.x version supported by the backend CodexAdapter.
    return MIN_CODEXADAPTER_OPENAI_AGENTS_VERSION


@lru_cache(maxsize=1)
def _detect_openai_agents_version() -> str | None:
    """Detect openai-agents package version for codex_tool-backed runtimes."""
    try:
        import agents

        module_version = getattr(agents, "__version__", None)
        if isinstance(module_version, str) and module_version:
            return module_version
    except ImportError:
        pass

    for package_name in ("openai-agents", "openai_agents"):
        try:
            return importlib.metadata.version(package_name)
        except importlib.metadata.PackageNotFoundError:
            continue
    return None


# SDK type constants
SDK_TYPE_CLAUDE = "claude"
SDK_TYPE_CODEX = "codex"
SDK_TYPE_TERMINAL_USE = "terminal_use"
DEFAULT_TERMINAL_USE_VERSION = "1.0.0"



# This file was auto-generated by Fern from our API Definition.

import datetime as dt
import typing

import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel


class State(UniversalBaseModel):
    """
    Represents a state in the agent system. A state is associated uniquely with a task and an agent.

    This entity is used to store states in MongoDB, with each state
    associated with a specific task and agent. The combination of task_id and agent_id is globally unique.

    The state is a dictionary of arbitrary data.
    """

    agent_id: str
    created_at: dt.datetime = pydantic.Field()
    """
    The timestamp when the state was created
    """

    id: str = pydantic.Field()
    """
    The task state's unique id
    """

    state: typing.Dict[str, typing.Any]
    task_id: str
    updated_at: typing.Optional[dt.datetime] = pydantic.Field(default=None)
    """
    The timestamp when the state was last updated
    """

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



from __future__ import annotations

import httpx
import pytest

from terminaluse.core.force_multipart import FORCE_MULTIPART
from terminaluse.core.http_client import AsyncHttpClient, HttpClient


class _FakeSyncHttpxClient:
    def __init__(self) -> None:
        self.calls: list[dict] = []

    def request(self, **kwargs):
        self.calls.append(kwargs)
        status_code = 500 if len(self.calls) == 1 else 200
        return httpx.Response(status_code=status_code)


class _FakeAsyncHttpxClient:
    def __init__(self) -> None:
        self.calls: list[dict] = []

    async def request(self, **kwargs):
        self.calls.append(kwargs)
        status_code = 500 if len(self.calls) == 1 else 200
        return httpx.Response(status_code=status_code)


def test_sync_retry_preserves_data_and_force_multipart():
    fake_httpx = _FakeSyncHttpxClient()
    client = HttpClient(
        httpx_client=fake_httpx,  # type: ignore[arg-type]
        base_timeout=lambda: None,
        base_headers=lambda: {},
    )

    response = client.request(
        method="POST",
        base_url="https://api.example.com",
        path="/v1/test",
        data={"name": "demo", "drop": None},
        force_multipart=True,
    )

    assert response.status_code == 200
    assert len(fake_httpx.calls) == 2
    assert fake_httpx.calls[0]["files"] == FORCE_MULTIPART
    assert fake_httpx.calls[1]["files"] == FORCE_MULTIPART
    assert fake_httpx.calls[0]["data"] == {"name": "demo"}
    assert fake_httpx.calls[1]["data"] == {"name": "demo"}


@pytest.mark.asyncio
async def test_async_retry_preserves_data_and_force_multipart():
    fake_httpx = _FakeAsyncHttpxClient()
    client = AsyncHttpClient(
        httpx_client=fake_httpx,  # type: ignore[arg-type]
        base_timeout=lambda: None,
        base_headers=lambda: {},
    )

    response = await client.request(
        method="POST",
        base_url="https://api.example.com",
        path="/v1/test",
        data={"name": "demo", "drop": None},
        force_multipart=True,
    )

    assert response.status_code == 200
    assert len(fake_httpx.calls) == 2
    assert fake_httpx.calls[0]["files"] == FORCE_MULTIPART
    assert fake_httpx.calls[1]["files"] == FORCE_MULTIPART
    assert fake_httpx.calls[0]["data"] == {"name": "demo"}
    assert fake_httpx.calls[1]["data"] == {"name": "demo"}



# ruff: noqa: ARG001
import asyncio
from pathlib import Path
from types import SimpleNamespace
from typing import Any
from unittest.mock import AsyncMock, Mock, patch

import pytest
from fastapi.testclient import TestClient

from terminaluse.lib.sandbox.runner import SandboxResult
from terminaluse.lib.sdk.fastacp.base.base_acp_server import (
    BaseACPServer,
    SandboxedHandlerError,
    UnsandboxableHandlerError,
)
from terminaluse.lib.types.acp import RPCMethod
from terminaluse.lib.types.task_context import TaskContext
from terminaluse.types.event import Event


async def _sandboxable_event_handler(ctx: TaskContext, event: Event):
    return {"ok": True}


class TestBaseACPServerInitialization:
    """Test BaseACPServer initialization and setup"""

    def test_base_acp_server_init(self):
        """Test BaseACPServer initialization sets up routes correctly"""
        with patch.dict("os.environ", {"TERMINALUSE_BASE_URL": ""}):
            server = BaseACPServer()

            # Check that FastAPI routes are set up
            routes = [route.path for route in server.routes]  # type: ignore[attr-defined]
            assert "/healthz" in routes
            assert "/api" in routes

            # Check that handlers dict is initialized
            assert hasattr(server, "_handlers")
            assert isinstance(server._handlers, dict)

    def test_base_acp_server_create_classmethod(self):
        """Test BaseACPServer.create() class method"""
        with patch.dict("os.environ", {"TERMINALUSE_BASE_URL": ""}):
            server = BaseACPServer.create()

            assert isinstance(server, BaseACPServer)
            assert hasattr(server, "_handlers")

    def test_lifespan_function_setup(self):
        """Test that lifespan function is properly configured"""
        with patch.dict("os.environ", {"TERMINALUSE_BASE_URL": ""}):
            server = BaseACPServer()

            # Check that lifespan is configured
            assert server.router.lifespan_context is not None

    def test_require_node_raises_when_node_missing(self):
        """TERMINALUSE_REQUIRE_NODE should fail fast if node is not sandbox-visible."""
        with patch.dict(
            "os.environ",
            {"TERMINALUSE_BASE_URL": "", "TERMINALUSE_REQUIRE_NODE": "true"},
            clear=False,
        ):
            server = BaseACPServer()
            with patch("terminaluse.lib.sdk.fastacp.base.base_acp_server.shutil.which", return_value=None):
                with pytest.raises(RuntimeError, match="TERMINALUSE_REQUIRE_NODE=true"):
                    server._enforce_required_runtime_binaries()

    def test_require_node_passes_when_node_present(self):
        """TERMINALUSE_REQUIRE_NODE should pass when node exists in sandbox path."""
        with patch.dict(
            "os.environ",
            {"TERMINALUSE_BASE_URL": "", "TERMINALUSE_REQUIRE_NODE": "true"},
            clear=False,
        ):
            server = BaseACPServer()
            with patch("terminaluse.lib.sdk.fastacp.base.base_acp_server.shutil.which", return_value="/usr/bin/node"):
                server._enforce_required_runtime_binaries()


class TestHealthCheckEndpoint:
    """Test health check endpoint functionality"""

    def test_health_check_endpoint(self, base_acp_server):
        """Test GET /healthz endpoint returns correct response"""
        client = TestClient(base_acp_server)

        response = client.get("/healthz")

        assert response.status_code == 200
        assert response.json() == {"status": "healthy"}

    def test_health_check_content_type(self, base_acp_server):
        """Test health check returns JSON content type"""
        client = TestClient(base_acp_server)

        response = client.get("/healthz")

        assert response.headers["content-type"] == "application/json"


class TestJSONRPCEndpointCore:
    """Test core JSON-RPC endpoint functionality"""

    def test_jsonrpc_endpoint_exists(self, base_acp_server):
        """Test POST /api endpoint exists"""
        client = TestClient(base_acp_server)

        # Send a basic request to check endpoint exists
        response = client.post("/api", json={})

        # Should not return 404 (endpoint exists)
        assert response.status_code != 404

    def test_jsonrpc_malformed_request(self, base_acp_server):
        """Test JSON-RPC endpoint handles malformed requests"""
        client = TestClient(base_acp_server)

        # Send malformed JSON
        response = client.post("/api", json={"invalid": "request"})

        assert response.status_code == 200
        data = response.json()
        assert "error" in data
        assert data["jsonrpc"] == "2.0"

    def test_jsonrpc_method_not_found(self, base_acp_server):
        """Test JSON-RPC method not found error"""
        client = TestClient(base_acp_server)

        request = {
            "jsonrpc": "2.0",
            "method": "nonexistent/method",
            "params": {},
            "id": "test-1",
        }

        response = client.post("/api", json=request)

        assert response.status_code == 200
        data = response.json()
        assert "error" in data
        assert data["error"]["code"] == -32601  # Method not found
        assert data["id"] == "test-1"

    def test_jsonrpc_valid_request_structure(self, base_acp_server):
        """Test JSON-RPC request parsing with valid structure"""
        client = TestClient(base_acp_server)

        # Add a mock handler for testing
        async def mock_handler(params):
            return {"status": "success"}

        base_acp_server._handlers[RPCMethod.EVENT_SEND] = mock_handler

        request = {
            "jsonrpc": "2.0",
            "method": "event/send",
            "params": {
                "agent": {
                    "id": "test-agent",
                    "name": "test-agent",
                    "namespace_id": "ns-123",
                    "description": "test agent",
                    "created_at": "2023-01-01T00:00:00Z",
                    "updated_at": "2023-01-01T00:00:00Z",
                },
                "task": {
                    "id": "test-task",
                    "namespace_id": "ns-123",
                    "filesystem_id": "fs-123",
                    "status": "RUNNING",
                },
                "event": {
                    "id": "event-123",
                    "agent_id": "test-agent",
                    "task_id": "test-task",
                    "sequence_id": 1,
                    "created_at": "2023-01-01T00:00:00Z",
                    "content": {
                        "type": "text",
                        "text": "test message",
                    },
                },
            },
            "id": "test-1",
        }

        response = client.post("/api", json=request)

        assert response.status_code == 200
        data = response.json()
        assert data["jsonrpc"] == "2.0"
        assert data["id"] == "test-1"
        # Should return immediate acknowledgment
        assert data["result"]["status"] == "processing"


class TestHandlerRegistration:
    """Test handler registration and management"""

    def test_on_event_decorator(self):
        """Test on_event decorator registration"""
        with patch.dict("os.environ", {"TERMINALUSE_BASE_URL": ""}):
            server = BaseACPServer()

            @server.on_event
            async def test_handler(ctx: TaskContext, event: Event):
                return {"test": "response"}

            # Check handler is registered
            assert RPCMethod.EVENT_SEND in server._handlers
            assert server._handlers[RPCMethod.EVENT_SEND] is not None

    def test_on_cancel_decorator(self):
        """Test on_cancel decorator registration"""
        with patch.dict("os.environ", {"TERMINALUSE_BASE_URL": ""}):
            server = BaseACPServer()

            @server.on_cancel
            async def test_handler(ctx: TaskContext):
                return {"test": "response"}

            # Check handler is registered
            assert RPCMethod.TASK_CANCEL in server._handlers
            assert server._handlers[RPCMethod.TASK_CANCEL] is not None

    def test_on_create_decorator(self):
        """Test on_create decorator registration"""
        with patch.dict("os.environ", {"TERMINALUSE_BASE_URL": ""}):
            server = BaseACPServer()

            @server.on_create
            async def test_handler(ctx: TaskContext, params: dict[str, Any]):
                return {"test": "response"}

            # Check handler is registered
            assert RPCMethod.TASK_CREATE in server._handlers
            assert server._handlers[RPCMethod.TASK_CREATE] is not None

    def test_on_create_registers_with_post_sync_policy(self):
        """task/create handlers must always register with post-sync enabled."""
        with patch.dict("os.environ", {"TERMINALUSE_BASE_URL": ""}):
            server = BaseACPServer()
            with patch.object(server, "_register_handler") as register_handler:

                @server.on_create
                async def test_handler(ctx: TaskContext, params: dict[str, Any]):
                    return {"test": "response"}

                assert test_handler is not None

            assert register_handler.call_count == 1
            assert register_handler.call_args.kwargs["sync_up_after"] is True

    def test_on_event_registers_with_post_sync_policy(self):
        """event/send handlers must always register with post-sync enabled."""
        with patch.dict("os.environ", {"TERMINALUSE_BASE_URL": ""}):
            server = BaseACPServer()
            with patch.object(server, "_register_handler") as register_handler:

                @server.on_event
                async def test_handler(ctx: TaskContext, event: Event):
                    return {"test": "response"}

                assert test_handler is not None

            assert register_handler.call_count == 1
            assert register_handler.call_args.kwargs["sync_up_after"] is True

    @pytest.mark.asyncio
    async def test_on_create_performs_sync_down_and_sync_up(self, sample_create_task_params):
        """Test on_create wrapper runs both pre-sync and post-sync operations."""
        with patch.dict("os.environ", {"TERMINALUSE_BASE_URL": ""}):
            server = BaseACPServer()
            local_fs_path = Path("/tmp/test-fs-123")

            filesystem_module = Mock()
            filesystem_module.get_filesystem_path = Mock(return_value=local_fs_path)
            filesystem_module.sync_down = AsyncMock(return_value=None)
            filesystem_module.sync_up = AsyncMock(return_value=None)
            task_module = Mock()
            task_module.sync_down_system_folder = AsyncMock(return_value=None)
            task_module.sync_up_system_folder = AsyncMock(return_value=None)

            server._filesystem_module = filesystem_module
            server._task_module = task_module

            handler_called = False

            @server.on_create
            async def test_handler(ctx: TaskContext, params: dict[str, Any]):
                nonlocal handler_called
                handler_called = True
                return {"ok": True}

            with (
                patch("terminaluse.lib.sdk.fastacp.base.base_acp_server.load_config", return_value={}),
                patch(
                    "terminaluse.lib.sdk.fastacp.base.base_acp_server.assemble_sandbox_mounts",
                    return_value=[],
                ) as assemble_sandbox_mounts_mock,
                patch("terminaluse.lib.sdk.fastacp.base.base_acp_server.write_sandbox_mounts_file"),
            ):
                await server._handlers[RPCMethod.TASK_CREATE](sample_create_task_params)

            filesystem_module.get_filesystem_path.assert_called_once_with(sample_create_task_params.task.filesystem_id)
            filesystem_module.sync_down.assert_awaited_once_with(
                sample_create_task_params.task.filesystem_id,
                local_path=local_fs_path,
            )
            task_module.sync_down_system_folder.assert_awaited_once_with(sample_create_task_params.task.id)
            filesystem_module.sync_up.assert_awaited_once_with(
                sample_create_task_params.task.filesystem_id,
                local_path=local_fs_path,
            )
            task_module.sync_up_system_folder.assert_awaited_once_with(sample_create_task_params.task.id)
            assemble_sandbox_mounts_mock.assert_called_once()
            call_args, call_kwargs = assemble_sandbox_mounts_mock.call_args
            assert call_args[0] == sample_create_task_params.task.id
            assert call_args[1] == sample_create_task_params.task.filesystem_id
            assert call_args[2] == {}
            assert isinstance(call_args[3], str)
            assert call_kwargs["filesystem_host_path"] == str(local_fs_path)
            assert call_kwargs["workspace_read_only"] is False
            assert handler_called is True

    @pytest.mark.asyncio
    async def test_on_create_uses_read_only_workspace_when_task_metadata_requests_it(self, sample_create_task_params):
        """filesystem_access_mode=ro should mount /workspace read-only."""
        with patch.dict("os.environ", {"TERMINALUSE_BASE_URL": ""}):
            server = BaseACPServer()
            local_fs_path = Path("/tmp/test-fs-123")

            filesystem_module = Mock()
            filesystem_module.get_filesystem_path = Mock(return_value=local_fs_path)
            filesystem_module.sync_down = AsyncMock(return_value=None)
            filesystem_module.sync_up = AsyncMock(return_value=None)
            task_module = Mock()
            task_module.sync_down_system_folder = AsyncMock(return_value=None)
            task_module.sync_up_system_folder = AsyncMock(return_value=None)
            server._filesystem_module = filesystem_module
            server._task_module = task_module

            @server.on_create
            async def test_handler(ctx: TaskContext, params: dict[str, Any]):
                return {"ok": True}

            ro_task = sample_create_task_params.task.model_copy(
                update={"task_metadata": {"filesystem_access_mode": "ro"}}
            )
            ro_params = sample_create_task_params.model_copy(update={"task": ro_task})

            with (
                patch("terminaluse.lib.sdk.fastacp.base.base_acp_server.load_config", return_value={}),
                patch(
                    "terminaluse.lib.sdk.fastacp.base.base_acp_server.assemble_sandbox_mounts",
                    return_value=[],
                ) as assemble_sandbox_mounts_mock,
                patch("terminaluse.lib.sdk.fastacp.base.base_acp_server.write_sandbox_mounts_file"),
            ):
                await server._handlers[RPCMethod.TASK_CREATE](ro_params)

            call_args, call_kwargs = assemble_sandbox_mounts_mock.call_args
            assert call_args[0] == ro_params.task.id
            assert call_args[1] == ro_params.task.filesystem_id
            assert call_kwargs["workspace_read_only"] is True

    @pytest.mark.asyncio
    async def test_on_create_syncs_up_even_when_handler_fails(self, sample_create_task_params):
        """Test on_create still performs post-sync when handler raises."""
        with patch.dict("os.environ", {"TERMINALUSE_BASE_URL": ""}):
            server = BaseACPServer()
            local_fs_path = Path("/tmp/test-fs-123")

            filesystem_module = Mock()
            filesystem_module.get_filesystem_path = Mock(return_value=local_fs_path)
            filesystem_module.sync_down = AsyncMock(return_value=None)
            filesystem_module.sync_up = AsyncMock(return_value=None)
            task_module = Mock()
            task_module.sync_down_system_folder = AsyncMock(return_value=None)
            task_module.sync_up_system_folder = AsyncMock(return_value=None)

            server._filesystem_module = filesystem_module
            server._task_module = task_module

            @server.on_create
            async def test_handler(ctx: TaskContext, params: dict[str, Any]):
                raise RuntimeError("handler failure")

            with (
                patch("terminaluse.lib.sdk.fastacp.base.base_acp_server.load_config", return_value={}),
                patch("terminaluse.lib.sdk.fastacp.base.base_acp_server.assemble_sandbox_mounts", return_value=[]),
                patch("terminaluse.lib.sdk.fastacp.base.base_acp_server.write_sandbox_mounts_file"),
                pytest.raises(RuntimeError, match="handler failure"),
            ):
                await server._handlers[RPCMethod.TASK_CREATE](sample_create_task_params)

            filesystem_module.get_filesystem_path.assert_called_once_with(sample_create_task_params.task.filesystem_id)
            filesystem_module.sync_down.assert_awaited_once_with(
                sample_create_task_params.task.filesystem_id,
                local_path=local_fs_path,
            )
            task_module.sync_down_system_folder.assert_awaited_once_with(sample_create_task_params.task.id)
            filesystem_module.sync_up.assert_awaited_once_with(
                sample_create_task_params.task.filesystem_id,
                local_path=local_fs_path,
            )
            task_module.sync_up_system_folder.assert_awaited_once_with(sample_create_task_params.task.id)


class TestSandboxRegistrationPolicy:
    """Test registration behavior for sandbox compatibility guarantees."""

    def test_nested_handler_fails_closed_by_default(self):
        with patch.dict(
            "os.environ",
            {"TERMINALUSE_BASE_URL": "", "TERMINALUSE_SANDBOX_FAIL_OPEN": "false"},
            clear=False,
        ):
            server = BaseACPServer()

            async def local_handler(ctx: TaskContext, event: Event):
                return {"ok": True}

            with pytest.raises(UnsandboxableHandlerError, match="module-level"):
                server.on_event(local_handler)

    def test_nested_handler_falls_back_when_fail_open_enabled(self):
        with patch.dict(
            "os.environ",
            {"TERMINALUSE_BASE_URL": "", "TERMINALUSE_SANDBOX_FAIL_OPEN": "true"},
            clear=False,
        ):
            server = BaseACPServer()

            async def local_handler(ctx: TaskContext, event: Event):
                return {"ok": True}

            server.on_event(local_handler)
            assert RPCMethod.EVENT_SEND in server._handlers
            assert RPCMethod.EVENT_SEND not in server._handler_refs

    def test_nested_handler_can_explicitly_allow_in_process(self):
        with patch.dict(
            "os.environ",
            {"TERMINALUSE_BASE_URL": "", "TERMINALUSE_SANDBOX_FAIL_OPEN": "false"},
            clear=False,
        ):
            server = BaseACPServer()

            @server.on_event(allow_in_process=True)
            async def local_handler(ctx: TaskContext, event: Event):
                return {"ok": True}

            assert local_handler is not None
            assert RPCMethod.EVENT_SEND in server._handlers
            assert RPCMethod.EVENT_SEND not in server._handler_refs

    def test_top_level_handler_registers_for_sandbox(self):
        with patch.dict(
            "os.environ",
            {"TERMINALUSE_BASE_URL": "", "TERMINALUSE_SANDBOX_FAIL_OPEN": "false"},
            clear=False,
        ):
            server = BaseACPServer()
            server.on_event(_sandboxable_event_handler)

            handler_ref = server._handler_refs[RPCMethod.EVENT_SEND]
            assert handler_ref.module == _sandboxable_event_handler.__module__
            assert handler_ref.function == _sandboxable_event_handler.__name__


class TestSandboxFailureHandling:
    """Test sandbox execution failure propagation and request logging behavior."""

    @pytest.mark.asyncio
    async def test_sandbox_failure_raises_sandboxed_handler_error(self):
        with patch.dict("os.environ", {"TERMINALUSE_BASE_URL": ""}):
            server = BaseACPServer()
            server.on_event(_sandboxable_event_handler)

            params = SimpleNamespace(task=SimpleNamespace(id=None, filesystem_id=None))
            failed_result = SandboxResult(
                success=False,
                exit_code=1,
                stdout='{"status":"error","error":"boom"}',
                stderr="trace",
                timed_out=False,
            )

            with patch(
                "terminaluse.lib.sdk.fastacp.base.base_acp_server.run_handler_sandboxed",
                new=AsyncMock(return_value=failed_result),
            ):
                with pytest.raises(SandboxedHandlerError, match="boom") as exc_info:
                    await server._handlers[RPCMethod.EVENT_SEND](params)

            assert exc_info.value.method == RPCMethod.EVENT_SEND
            assert exc_info.value.exit_code == 1
            assert exc_info.value.timed_out is False

    @pytest.mark.asyncio
    async def test_process_request_logs_error_not_success_on_sandbox_failure(self, caplog):
        with patch.dict("os.environ", {"TERMINALUSE_BASE_URL": ""}):
            server = BaseACPServer()
            server.on_event(_sandboxable_event_handler)

            params = SimpleNamespace(task=SimpleNamespace(id=None, filesystem_id=None))
            failed_result = SandboxResult(
                success=False,
                exit_code=1,
                stdout='{"status":"error","error":"boom"}',
                stderr="trace",
                timed_out=False,
            )

            caplog.set_level("INFO")
            with patch(
                "terminaluse.lib.sdk.fastacp.base.base_acp_server.run_handler_sandboxed",
                new=AsyncMock(return_value=failed_result),
            ):
                await server._process_request("req-1", RPCMethod.EVENT_SEND, params)

            messages = [record.getMessage() for record in caplog.records]
            assert any("Error processing request req-1" in message for message in messages)
            assert not any("Successfully processed request req-1" in message for message in messages)


class TestBackgroundProcessing:
    """Test background processing functionality"""

    @pytest.mark.asyncio
    async def test_notification_processing(self, async_base_acp_server):
        """Test notification processing (requests with no ID)"""
        # Add a mock handler
        handler_called = False
        received_params = None

        async def mock_handler(params):
            nonlocal handler_called, received_params
            handler_called = True
            received_params = params
            return {"status": "processed"}

        async_base_acp_server._handlers[RPCMethod.EVENT_SEND] = mock_handler

        client = TestClient(async_base_acp_server)

        request = {
            "jsonrpc": "2.0",
            "method": "event/send",
            "params": {
                "agent": {
                    "id": "test-agent",
                    "name": "test-agent",
                    "namespace_id": "ns-123",
                    "description": "test agent",
                    "created_at": "2023-01-01T00:00:00Z",
                    "updated_at": "2023-01-01T00:00:00Z",
                },
                "task": {
                    "id": "test-task",
                    "namespace_id": "ns-123",
                    "filesystem_id": "fs-123",
                    "status": "RUNNING",
                },
                "event": {
                    "id": "event-123",
                    "agent_id": "test-agent",
                    "task_id": "test-task",
                    "sequence_id": 1,
                    "created_at": "2023-01-01T00:00:00Z",
                    "content": {
                        "type": "text",
                        "text": "test message",
                    },
                },
            },
            # No ID = notification
        }

        response = client.post("/api", json=request)

        assert response.status_code == 200
        data = response.json()
        assert data["id"] is None  # Notification response

        # Give background task time to execute
        await asyncio.sleep(0.1)

        # Handler should have been called
        assert handler_called is True
        assert received_params is not None

    @pytest.mark.asyncio
    async def test_request_processing_with_id(self, async_base_acp_server):
        """Test request processing with ID returns immediate acknowledgment"""

        # Add a mock handler
        async def mock_handler(params):
            return {"status": "processed"}

        async_base_acp_server._handlers[RPCMethod.TASK_CANCEL] = mock_handler

        client = TestClient(async_base_acp_server)

        request = {
            "jsonrpc": "2.0",
            "method": "task/cancel",
            "params": {
                "agent": {
                    "id": "test-agent",
                    "name": "test-agent",
                    "namespace_id": "ns-123",
                    "description": "test agent",
                    "created_at": "2023-01-01T00:00:00Z",
                    "updated_at": "2023-01-01T00:00:00Z",
                },
                "task": {
                    "id": "test-task-123",
                    "namespace_id": "ns-123",
                    "filesystem_id": "fs-123",
                    "status": "RUNNING",
                },
            },
            "id": "test-request-1",
        }

        response = client.post("/api", json=request)

        assert response.status_code == 200
        data = response.json()
        assert data["jsonrpc"] == "2.0"
        assert data["id"] == "test-request-1"
        assert data["result"]["status"] == "processing"  # Immediate acknowledgment


class TestAsyncRPCMethods:
    """Test RPC methods that return processing status (async behavior)"""

    def test_create_task_async_response(self, base_acp_server):
        """Test that TASK_CREATE method returns processing status (async behavior)"""
        client = TestClient(base_acp_server)

        # Add a mock handler for init task
        async def mock_init_handler(params):
            return {
                "task_id": params.task.id,
                "status": "initialized",
            }

        base_acp_server._handlers[RPCMethod.TASK_CREATE] = mock_init_handler

        request = {
            "jsonrpc": "2.0",
            "method": "task/create",
            "params": {
                "agent": {
                    "id": "test-agent",
                    "name": "test-agent",
                    "namespace_id": "ns-123",
                    "description": "test agent",
                    "created_at": "2023-01-01T00:00:00Z",
                    "updated_at": "2023-01-01T00:00:00Z",
                },
                "task": {
                    "id": "test-task-456",
                    "namespace_id": "ns-123",
                    "filesystem_id": "fs-123",
                    "status": "RUNNING",
                },
            },
            "id": "test-init-1",
        }

        response = client.post("/api", json=request)

        assert response.status_code == 200
        data = response.json()

        # Verify JSON-RPC structure
        assert data["jsonrpc"] == "2.0"
        assert data["id"] == "test-init-1"
        assert "result" in data
        assert data.get("error") is None

        # Verify it returns async "processing" status (not the handler's result)
        result = data["result"]
        assert result["status"] == "processing"

        # Verify it's NOT the handler's actual result
        assert result.get("status") != "initialized"


class TestErrorHandling:
    """Test error handling scenarios"""

    def test_invalid_json_request(self, base_acp_server):
        """Test handling of invalid JSON in request body"""
        client = TestClient(base_acp_server)

        # Send invalid JSON
        response = client.post("/api", content="invalid json", headers={"Content-Type": "application/json"})

        assert response.status_code == 200
        data = response.json()
        assert "error" in data
        assert data["jsonrpc"] == "2.0"

    def test_missing_required_fields(self, base_acp_server):
        """Test handling of requests missing required JSON-RPC fields"""
        client = TestClient(base_acp_server)

        # Missing method field
        request = {"jsonrpc": "2.0", "params": {}, "id": "test-1"}

        response = client.post("/api", json=request)

        assert response.status_code == 200
        data = response.json()
        assert "error" in data

    def test_invalid_method_enum(self, base_acp_server):
        """Test handling of invalid method names"""
        client = TestClient(base_acp_server)

        request = {
            "jsonrpc": "2.0",
            "method": "invalid/method/name",
            "params": {},
            "id": "test-1",
        }

        response = client.post("/api", json=request)

        assert response.status_code == 200
        data = response.json()
        assert "error" in data
        assert data["error"]["code"] == -32601  # Method not found

    @pytest.mark.asyncio
    async def test_handler_exception_handling(self, async_base_acp_server):
        """Test that handler exceptions are properly handled"""

        # Add a handler that raises an exception
        async def failing_handler(params):
            raise ValueError("Test exception")

        async_base_acp_server._handlers[RPCMethod.EVENT_SEND] = failing_handler

        client = TestClient(async_base_acp_server)

        request = {
            "jsonrpc": "2.0",
            "method": "event/send",
            "params": {
                "agent": {
                    "id": "test-agent",
                    "name": "test-agent",
                    "namespace_id": "ns-123",
                    "description": "test agent",
                    "created_at": "2023-01-01T00:00:00Z",
                    "updated_at": "2023-01-01T00:00:00Z",
                },
                "task": {
                    "id": "test-task",
                    "namespace_id": "ns-123",
                    "filesystem_id": "fs-123",
                    "status": "RUNNING",
                },
                "event": {
                    "id": "event-123",
                    "agent_id": "test-agent",
                    "task_id": "test-task",
                    "sequence_id": 1,
                    "created_at": "2023-01-01T00:00:00Z",
                    "content": {
                        "type": "text",
                        "text": "test message",
                    },
                },
            },
            "id": "test-1",
        }

        response = client.post("/api", json=request)

        # Should still return immediate acknowledgment
        assert response.status_code == 200
        data = response.json()
        assert data["result"]["status"] == "processing"

        # Give background task time to fail
        await asyncio.sleep(0.1)
        # Exception should be logged but not crash the server


class TestHandlerEndSignaling:
    """Tests for handler-end signaling after event handler completes.

    After an EVENT_SEND handler finishes (success or failure), the agent server
    must call handler_end() to transition the task from RUNNING back to IDLE.
    Without this, tasks get permanently stuck in RUNNING.
    """

    @pytest.fixture
    def event_send_params(self):
        """Parsed SendEventParams-like object with task.id for handler-end tests."""
        return SimpleNamespace(
            task=SimpleNamespace(id="task-abc-123", filesystem_id="fs-123"),
            agent=SimpleNamespace(id="agent-1", name="test-agent", namespace_id="ns-1"),
            event=SimpleNamespace(id="evt-1", agent_id="agent-1", task_id="task-abc-123"),
        )

    @pytest.fixture
    def cancel_params(self):
        """Parsed CancelTaskParams-like object for non-event methods."""
        return SimpleNamespace(
            task=SimpleNamespace(id="task-abc-123", filesystem_id="fs-123"),
            agent=SimpleNamespace(id="agent-1", name="test-agent", namespace_id="ns-1"),
        )

    @pytest.mark.asyncio
    async def test_process_request_calls_handler_end_on_success(self, event_send_params):
        """After a successful EVENT_SEND handler, handler_end is called with success=True."""
        with patch.dict("os.environ", {"TERMINALUSE_BASE_URL": ""}):
            server = BaseACPServer()

            async def mock_handler(params):
                return {"ok": True}

            server._handlers[RPCMethod.EVENT_SEND] = mock_handler

            mock_client = AsyncMock()
            mock_client.tasks.handler_end = AsyncMock()
            server._terminaluse_client = mock_client

            await server._process_request("req-1", RPCMethod.EVENT_SEND, event_send_params)

            mock_client.tasks.handler_end.assert_awaited_once()
            call_kwargs = mock_client.tasks.handler_end.call_args
            assert call_kwargs.kwargs["task_id"] == "task-abc-123"
            assert call_kwargs.kwargs["event_id"] == "evt-1"
            assert call_kwargs.kwargs["success"] is True
            assert call_kwargs.kwargs["error"] is None
            assert isinstance(call_kwargs.kwargs["duration_ms"], int)

    @pytest.mark.asyncio
    async def test_process_request_calls_handler_end_on_failure(self, event_send_params):
        """After a failed EVENT_SEND handler, handler_end is called with success=False."""
        with patch.dict("os.environ", {"TERMINALUSE_BASE_URL": ""}):
            server = BaseACPServer()

            async def failing_handler(params):
                raise RuntimeError("handler exploded")

            server._handlers[RPCMethod.EVENT_SEND] = failing_handler

            mock_client = AsyncMock()
            mock_client.tasks.handler_end = AsyncMock()
            server._terminaluse_client = mock_client

            await server._process_request("req-2", RPCMethod.EVENT_SEND, event_send_params)

            mock_client.tasks.handler_end.assert_awaited_once()
            call_kwargs = mock_client.tasks.handler_end.call_args
            assert call_kwargs.kwargs["task_id"] == "task-abc-123"
            assert call_kwargs.kwargs["event_id"] == "evt-1"
            assert call_kwargs.kwargs["success"] is False
            assert "handler exploded" in call_kwargs.kwargs["error"]

    @pytest.mark.asyncio
    async def test_process_notification_calls_handler_end_on_success(self, event_send_params):
        """Notifications (no request ID) also call handler_end after EVENT_SEND."""
        with patch.dict("os.environ", {"TERMINALUSE_BASE_URL": ""}):
            server = BaseACPServer()

            async def mock_handler(params):
                return {"ok": True}

            server._handlers[RPCMethod.EVENT_SEND] = mock_handler

            mock_client = AsyncMock()
            mock_client.tasks.handler_end = AsyncMock()
            server._terminaluse_client = mock_client

            await server._process_notification(RPCMethod.EVENT_SEND, event_send_params)

            mock_client.tasks.handler_end.assert_awaited_once()
            call_kwargs = mock_client.tasks.handler_end.call_args
            assert call_kwargs.kwargs["task_id"] == "task-abc-123"
            assert call_kwargs.kwargs["event_id"] == "evt-1"
            assert call_kwargs.kwargs["success"] is True

    @pytest.mark.asyncio
    async def test_process_notification_calls_handler_end_on_failure(self, event_send_params):
        """Notifications call handler_end with success=False when handler fails."""
        with patch.dict("os.environ", {"TERMINALUSE_BASE_URL": ""}):
            server = BaseACPServer()

            async def failing_handler(params):
                raise ValueError("bad event")

            server._handlers[RPCMethod.EVENT_SEND] = failing_handler

            mock_client = AsyncMock()
            mock_client.tasks.handler_end = AsyncMock()
            server._terminaluse_client = mock_client

            await server._process_notification(RPCMethod.EVENT_SEND, event_send_params)

            mock_client.tasks.handler_end.assert_awaited_once()
            call_kwargs = mock_client.tasks.handler_end.call_args
            assert call_kwargs.kwargs["event_id"] == "evt-1"
            assert call_kwargs.kwargs["success"] is False
            assert "bad event" in call_kwargs.kwargs["error"]

    @pytest.mark.asyncio
    async def test_handler_end_omits_event_id_when_event_missing(self):
        """When params.event is missing, handler_end receives event_id=None."""
        with patch.dict("os.environ", {"TERMINALUSE_BASE_URL": ""}):
            server = BaseACPServer()

            async def mock_handler(params):
                return {"ok": True}

            server._handlers[RPCMethod.EVENT_SEND] = mock_handler

            params_no_event = SimpleNamespace(
                task=SimpleNamespace(id="task-abc-123", filesystem_id="fs-123"),
                agent=SimpleNamespace(id="agent-1", name="test-agent", namespace_id="ns-1"),
                event=None,
            )

            mock_client = AsyncMock()
            mock_client.tasks.handler_end = AsyncMock()
            server._terminaluse_client = mock_client

            await server._process_request("req-no-event", RPCMethod.EVENT_SEND, params_no_event)

            mock_client.tasks.handler_end.assert_awaited_once()
            call_kwargs = mock_client.tasks.handler_end.call_args
            assert call_kwargs.kwargs["task_id"] == "task-abc-123"
            assert call_kwargs.kwargs["event_id"] is None

    @pytest.mark.asyncio
    async def test_no_handler_end_for_non_event_methods(self, cancel_params):
        """handler_end should NOT be called for TASK_CANCEL or other non-event methods."""
        with patch.dict("os.environ", {"TERMINALUSE_BASE_URL": ""}):
            server = BaseACPServer()

            async def mock_handler(params):
                return {"ok": True}

            server._handlers[RPCMethod.TASK_CANCEL] = mock_handler

            mock_client = AsyncMock()
            mock_client.tasks.handler_end = AsyncMock()
            server._terminaluse_client = mock_client

            await server._process_request("req-3", RPCMethod.TASK_CANCEL, cancel_params)

            mock_client.tasks.handler_end.assert_not_awaited()

    @pytest.mark.asyncio
    async def test_handler_end_failure_does_not_crash_server(self, event_send_params):
        """If handler_end itself fails, the error is logged but doesn't propagate."""
        with patch.dict("os.environ", {"TERMINALUSE_BASE_URL": ""}):
            server = BaseACPServer()

            async def mock_handler(params):
                return {"ok": True}

            server._handlers[RPCMethod.EVENT_SEND] = mock_handler

            mock_client = AsyncMock()
            mock_client.tasks.handler_end = AsyncMock(side_effect=Exception("network error calling handler-end"))
            server._terminaluse_client = mock_client

            # Should not raise
            await server._process_request("req-4", RPCMethod.EVENT_SEND, event_send_params)

            mock_client.tasks.handler_end.assert_awaited_once()

    @pytest.mark.asyncio
    async def test_handler_end_includes_duration_ms(self, event_send_params):
        """handler_end should report handler duration in milliseconds."""
        with patch.dict("os.environ", {"TERMINALUSE_BASE_URL": ""}):
            server = BaseACPServer()

            async def slow_handler(params):
                await asyncio.sleep(0.05)  # 50ms
                return {"ok": True}

            server._handlers[RPCMethod.EVENT_SEND] = slow_handler

            mock_client = AsyncMock()
            mock_client.tasks.handler_end = AsyncMock()
            server._terminaluse_client = mock_client

            await server._process_request("req-5", RPCMethod.EVENT_SEND, event_send_params)

            call_kwargs = mock_client.tasks.handler_end.call_args
            duration_ms = call_kwargs.kwargs["duration_ms"]
            assert duration_ms >= 40  # At least ~40ms (allowing for timing variance)

    @pytest.mark.asyncio
    async def test_handler_end_not_called_when_no_task_id(self):
        """If params has no task.id, handler_end should be skipped gracefully."""
        with patch.dict("os.environ", {"TERMINALUSE_BASE_URL": ""}):
            server = BaseACPServer()

            async def mock_handler(params):
                return {"ok": True}

            server._handlers[RPCMethod.EVENT_SEND] = mock_handler

            mock_client = AsyncMock()
            mock_client.tasks.handler_end = AsyncMock()
            server._terminaluse_client = mock_client

            params_no_task = SimpleNamespace(task=SimpleNamespace(id=None, filesystem_id=None))
            await server._process_request("req-6", RPCMethod.EVENT_SEND, params_no_task)

            mock_client.tasks.handler_end.assert_not_awaited()

    @pytest.mark.asyncio
    async def test_no_handler_end_for_task_create(self):
        """handler_end should NOT be called for TASK_CREATE."""
        with patch.dict("os.environ", {"TERMINALUSE_BASE_URL": ""}):
            server = BaseACPServer()

            async def mock_handler(params):
                return {"ok": True}

            server._handlers[RPCMethod.TASK_CREATE] = mock_handler

            mock_client = AsyncMock()
            mock_client.tasks.handler_end = AsyncMock()
            server._terminaluse_client = mock_client

            create_params = SimpleNamespace(
                task=SimpleNamespace(id="task-abc-123", filesystem_id="fs-123"),
                agent=SimpleNamespace(id="agent-1", name="test-agent", namespace_id="ns-1"),
            )
            await server._process_request("req-7", RPCMethod.TASK_CREATE, create_params)

            mock_client.tasks.handler_end.assert_not_awaited()

    @pytest.mark.asyncio
    async def test_client_creation_failure_does_not_crash(self, event_send_params):
        """If _get_terminaluse_client() throws, the error is swallowed."""
        with patch.dict("os.environ", {"TERMINALUSE_BASE_URL": ""}):
            server = BaseACPServer()

            async def mock_handler(params):
                return {"ok": True}

            server._handlers[RPCMethod.EVENT_SEND] = mock_handler

            # Force client creation to fail — no pre-set _terminaluse_client
            server._terminaluse_client = None
            with patch.object(
                server,
                "_get_terminaluse_client",
                side_effect=RuntimeError("env vars missing"),
            ):
                # Should not raise
                await server._process_request("req-8", RPCMethod.EVENT_SEND, event_send_params)

    @pytest.mark.asyncio
    async def test_handler_crash_triggers_handler_end_with_error_details(self, event_send_params):
        """Full round-trip: handler crash → handler_end called with failure info → server survives.

        Proves that a crashed handler doesn't leave the task stuck in RUNNING:
        the SDK signals handler_end(success=False) so Nucleus can transition back to IDLE.
        """
        with patch.dict("os.environ", {"TERMINALUSE_BASE_URL": ""}):
            server = BaseACPServer()

            async def crashing_handler(params):
                raise RuntimeError("segfault in user code")

            server._handlers[RPCMethod.EVENT_SEND] = crashing_handler

            mock_client = AsyncMock()
            mock_client.tasks.handler_end = AsyncMock()
            server._terminaluse_client = mock_client

            # Process via notification (fire-and-forget path, like real deployments)
            await server._process_notification(RPCMethod.EVENT_SEND, event_send_params)

            # handler_end was called exactly once
            mock_client.tasks.handler_end.assert_awaited_once()
            call_kwargs = mock_client.tasks.handler_end.call_args.kwargs
            assert call_kwargs["task_id"] == "task-abc-123"
            assert call_kwargs["success"] is False
            assert "segfault in user code" in call_kwargs["error"]
            assert isinstance(call_kwargs["duration_ms"], int)
            assert call_kwargs["duration_ms"] >= 0



"""
Debug configuration models for TerminalUse CLI debugging.
"""

import socket
from enum import Enum

from terminaluse.lib.utils.model_utils import BaseModel


class DebugMode(str, Enum):
    """Debug mode options"""

    WORKER = "worker"
    ACP = "acp"
    BOTH = "both"
    NONE = "none"


class DebugConfig(BaseModel):
    """Configuration for debug mode"""

    enabled: bool = False
    mode: DebugMode = DebugMode.NONE
    port: int = 5678
    wait_for_attach: bool = False
    auto_port: bool = True  # Automatically find available port if specified port is busy

    @classmethod
    def create_worker_debug(
        cls, port: int = 5678, wait_for_attach: bool = False, auto_port: bool = True
    ) -> "DebugConfig":
        """Create debug config for worker debugging"""
        return cls(enabled=True, mode=DebugMode.WORKER, port=port, wait_for_attach=wait_for_attach, auto_port=auto_port)

    @classmethod
    def create_acp_debug(cls, port: int = 5679, wait_for_attach: bool = False, auto_port: bool = True) -> "DebugConfig":
        """Create debug config for ACP debugging"""
        return cls(enabled=True, mode=DebugMode.ACP, port=port, wait_for_attach=wait_for_attach, auto_port=auto_port)

    @classmethod
    def create_both_debug(
        cls, worker_port: int = 5678, _acp_port: int = 5679, wait_for_attach: bool = False, auto_port: bool = True
    ) -> "DebugConfig":
        """Create debug config for both worker and ACP debugging"""
        return cls(
            enabled=True,
            mode=DebugMode.BOTH,
            port=worker_port,  # Primary port for worker
            wait_for_attach=wait_for_attach,
            auto_port=auto_port,
        )

    def should_debug_worker(self) -> bool:
        """Check if worker should be debugged"""
        return self.enabled and self.mode in (DebugMode.WORKER, DebugMode.BOTH)

    def should_debug_acp(self) -> bool:
        """Check if ACP should be debugged"""
        return self.enabled and self.mode in (DebugMode.ACP, DebugMode.BOTH)

    def get_worker_port(self) -> int:
        """Get port for worker debugging"""
        return self.port

    def get_acp_port(self) -> int:
        """Get port for ACP debugging"""
        if self.mode == DebugMode.BOTH:
            return self.port + 1  # Use port + 1 for ACP when debugging both
        return self.port


def find_available_port(start_port: int = 5678, max_attempts: int = 10) -> int:
    """Find an available port starting from start_port"""
    for port in range(start_port, start_port + max_attempts):
        try:
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                s.bind(("localhost", port))
                return port
        except OSError:
            continue

    # If we can't find an available port, just return the start port
    # and let the debug server handle the error
    return start_port


def resolve_debug_port(config: DebugConfig, target_port: int) -> int:
    """Resolve the actual port to use for debugging"""
    if config.auto_port:
        return find_available_port(target_port)
    return target_port



# This file was auto-generated by Fern from our API Definition.

import typing

import pydantic
import typing_extensions
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel
from ..core.serialization import FieldMetadata


class Usage(UniversalBaseModel):
    """
    Token usage statistics.
    """

    input_tokens: typing_extensions.Annotated[int, FieldMetadata(alias="inputTokens")] = pydantic.Field(
        alias="inputTokens"
    )
    """
    Input tokens used
    """

    output_tokens: typing_extensions.Annotated[int, FieldMetadata(alias="outputTokens")] = pydantic.Field(
        alias="outputTokens"
    )
    """
    Output tokens used
    """

    total_tokens: typing_extensions.Annotated[typing.Optional[int], FieldMetadata(alias="totalTokens")] = (
        pydantic.Field(alias="totalTokens", default=None)
    )
    """
    Total tokens (input + output)
    """

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



# This file was auto-generated by Fern from our API Definition.

import typing

from ..core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ..core.request_options import RequestOptions
from ..types.delete_response import DeleteResponse
from ..types.project import Project
from ..types.project_collaborator_response import ProjectCollaboratorResponse
from ..types.project_collaborators_response import ProjectCollaboratorsResponse
from ..types.project_list_item import ProjectListItem
from .raw_client import AsyncRawProjectsClient, RawProjectsClient
from .types.upsert_project_collaborator_request_role import UpsertProjectCollaboratorRequestRole

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class ProjectsClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._raw_client = RawProjectsClient(client_wrapper=client_wrapper)

    @property
    def with_raw_response(self) -> RawProjectsClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        RawProjectsClient
        """
        return self._raw_client

    def list(
        self,
        *,
        namespace_id: typing.Optional[str] = None,
        limit: typing.Optional[int] = None,
        page_number: typing.Optional[int] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> typing.List[ProjectListItem]:
        """
        List all projects the user has access to.

        Parameters
        ----------
        namespace_id : typing.Optional[str]
            Filter by namespace ID

        limit : typing.Optional[int]
            Maximum number of results

        page_number : typing.Optional[int]
            Page number

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        typing.List[ProjectListItem]
            Successful Response

        Examples
        --------
        from terminaluse import TerminalUseBase

        client = TerminalUseBase(
            agent_api_key="YOUR_AGENT_API_KEY",
            token="YOUR_TOKEN",
            base_url="https://yourhost.com/path/to/api",
        )
        client.projects.list()
        """
        _response = self._raw_client.list(
            namespace_id=namespace_id, limit=limit, page_number=page_number, request_options=request_options
        )
        return _response.data

    def create(
        self,
        *,
        name: str,
        namespace_id: str,
        description: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> Project:
        """
        Create a new project within a namespace.

        Parameters
        ----------
        name : str
            Project name (unique within namespace).

        namespace_id : str
            The namespace this project belongs to.

        description : typing.Optional[str]
            Optional project description.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        Project
            Successful Response

        Examples
        --------
        from terminaluse import TerminalUseBase

        client = TerminalUseBase(
            agent_api_key="YOUR_AGENT_API_KEY",
            token="YOUR_TOKEN",
            base_url="https://yourhost.com/path/to/api",
        )
        client.projects.create(
            name="name",
            namespace_id="namespace_id",
        )
        """
        _response = self._raw_client.create(
            name=name, namespace_id=namespace_id, description=description, request_options=request_options
        )
        return _response.data

    def retrieve(self, project_id: str, *, request_options: typing.Optional[RequestOptions] = None) -> Project:
        """
        Get a project by ID.

        Parameters
        ----------
        project_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        Project
            Successful Response

        Examples
        --------
        from terminaluse import TerminalUseBase

        client = TerminalUseBase(
            agent_api_key="YOUR_AGENT_API_KEY",
            token="YOUR_TOKEN",
            base_url="https://yourhost.com/path/to/api",
        )
        client.projects.retrieve(
            project_id="project_id",
        )
        """
        _response = self._raw_client.retrieve(project_id, request_options=request_options)
        return _response.data

    def delete(self, project_id: str, *, request_options: typing.Optional[RequestOptions] = None) -> DeleteResponse:
        """
        Delete a project (must be empty).

        Parameters
        ----------
        project_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        DeleteResponse
            Successful Response

        Examples
        --------
        from terminaluse import TerminalUseBase

        client = TerminalUseBase(
            agent_api_key="YOUR_AGENT_API_KEY",
            token="YOUR_TOKEN",
            base_url="https://yourhost.com/path/to/api",
        )
        client.projects.delete(
            project_id="project_id",
        )
        """
        _response = self._raw_client.delete(project_id, request_options=request_options)
        return _response.data

    def update(
        self,
        project_id: str,
        *,
        description: typing.Optional[str] = OMIT,
        name: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> Project:
        """
        Update a project's properties.

        Parameters
        ----------
        project_id : str

        description : typing.Optional[str]
            Optional project description.

        name : typing.Optional[str]
            Project name.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        Project
            Successful Response

        Examples
        --------
        from terminaluse import TerminalUseBase

        client = TerminalUseBase(
            agent_api_key="YOUR_AGENT_API_KEY",
            token="YOUR_TOKEN",
            base_url="https://yourhost.com/path/to/api",
        )
        client.projects.update(
            project_id="project_id",
        )
        """
        _response = self._raw_client.update(
            project_id, description=description, name=name, request_options=request_options
        )
        return _response.data

    def list_collaborators(
        self, project_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> ProjectCollaboratorsResponse:
        """
        List explicit collaborator roles on a project.

        Parameters
        ----------
        project_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ProjectCollaboratorsResponse
            Successful Response

        Examples
        --------
        from terminaluse import TerminalUseBase

        client = TerminalUseBase(
            agent_api_key="YOUR_AGENT_API_KEY",
            token="YOUR_TOKEN",
            base_url="https://yourhost.com/path/to/api",
        )
        client.projects.list_collaborators(
            project_id="project_id",
        )
        """
        _response = self._raw_client.list_collaborators(project_id, request_options=request_options)
        return _response.data

    def upsert_collaborator(
        self,
        project_id: str,
        member_id: str,
        *,
        role: UpsertProjectCollaboratorRequestRole,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> ProjectCollaboratorResponse:
        """
        Create or update an explicit collaborator role on a project.

        Parameters
        ----------
        project_id : str

        member_id : str

        role : UpsertProjectCollaboratorRequestRole
            Role to assign.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ProjectCollaboratorResponse
            Successful Response

        Examples
        --------
        from terminaluse import TerminalUseBase

        client = TerminalUseBase(
            agent_api_key="YOUR_AGENT_API_KEY",
            token="YOUR_TOKEN",
            base_url="https://yourhost.com/path/to/api",
        )
        client.projects.upsert_collaborator(
            project_id="project_id",
            member_id="member_id",
            role="discoverer",
        )
        """
        _response = self._raw_client.upsert_collaborator(
            project_id, member_id, role=role, request_options=request_options
        )
        return _response.data

    def remove_collaborator(
        self, project_id: str, member_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> DeleteResponse:
        """
        Remove explicit collaborator role tuples from a project.

        Parameters
        ----------
        project_id : str

        member_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        DeleteResponse
            Successful Response

        Examples
        --------
        from terminaluse import TerminalUseBase

        client = TerminalUseBase(
            agent_api_key="YOUR_AGENT_API_KEY",
            token="YOUR_TOKEN",
            base_url="https://yourhost.com/path/to/api",
        )
        client.projects.remove_collaborator(
            project_id="project_id",
            member_id="member_id",
        )
        """
        _response = self._raw_client.remove_collaborator(project_id, member_id, request_options=request_options)
        return _response.data


class AsyncProjectsClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._raw_client = AsyncRawProjectsClient(client_wrapper=client_wrapper)

    @property
    def with_raw_response(self) -> AsyncRawProjectsClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        AsyncRawProjectsClient
        """
        return self._raw_client

    async def list(
        self,
        *,
        namespace_id: typing.Optional[str] = None,
        limit: typing.Optional[int] = None,
        page_number: typing.Optional[int] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> typing.List[ProjectListItem]:
        """
        List all projects the user has access to.

        Parameters
        ----------
        namespace_id : typing.Optional[str]
            Filter by namespace ID

        limit : typing.Optional[int]
            Maximum number of results

        page_number : typing.Optional[int]
            Page number

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        typing.List[ProjectListItem]
            Successful Response

        Examples
        --------
        import asyncio

        from terminaluse import AsyncTerminalUseBase

        client = AsyncTerminalUseBase(
            agent_api_key="YOUR_AGENT_API_KEY",
            token="YOUR_TOKEN",
            base_url="https://yourhost.com/path/to/api",
        )


        async def main() -> None:
            await client.projects.list()


        asyncio.run(main())
        """
        _response = await self._raw_client.list(
            namespace_id=namespace_id, limit=limit, page_number=page_number, request_options=request_options
        )
        return _response.data

    async def create(
        self,
        *,
        name: str,
        namespace_id: str,
        description: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> Project:
        """
        Create a new project within a namespace.

        Parameters
        ----------
        name : str
            Project name (unique within namespace).

        namespace_id : str
            The namespace this project belongs to.

        description : typing.Optional[str]
            Optional project description.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        Project
            Successful Response

        Examples
        --------
        import asyncio

        from terminaluse import AsyncTerminalUseBase

        client = AsyncTerminalUseBase(
            agent_api_key="YOUR_AGENT_API_KEY",
            token="YOUR_TOKEN",
            base_url="https://yourhost.com/path/to/api",
        )


        async def main() -> None:
            await client.projects.create(
                name="name",
                namespace_id="namespace_id",
            )


        asyncio.run(main())
        """
        _response = await self._raw_client.create(
            name=name, namespace_id=namespace_id, description=description, request_options=request_options
        )
        return _response.data

    async def retrieve(self, project_id: str, *, request_options: typing.Optional[RequestOptions] = None) -> Project:
        """
        Get a project by ID.

        Parameters
        ----------
        project_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        Project
            Successful Response

        Examples
        --------
        import asyncio

        from terminaluse import AsyncTerminalUseBase

        client = AsyncTerminalUseBase(
            agent_api_key="YOUR_AGENT_API_KEY",
            token="YOUR_TOKEN",
            base_url="https://yourhost.com/path/to/api",
        )


        async def main() -> None:
            await client.projects.retrieve(
                project_id="project_id",
            )


        asyncio.run(main())
        """
        _response = await self._raw_client.retrieve(project_id, request_options=request_options)
        return _response.data

    async def delete(
        self, project_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> DeleteResponse:
        """
        Delete a project (must be empty).

        Parameters
        ----------
        project_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        DeleteResponse
            Successful Response

        Examples
        --------
        import asyncio

        from terminaluse import AsyncTerminalUseBase

        client = AsyncTerminalUseBase(
            agent_api_key="YOUR_AGENT_API_KEY",
            token="YOUR_TOKEN",
            base_url="https://yourhost.com/path/to/api",
        )


        async def main() -> None:
            await client.projects.delete(
                project_id="project_id",
            )


        asyncio.run(main())
        """
        _response = await self._raw_client.delete(project_id, request_options=request_options)
        return _response.data

    async def update(
        self,
        project_id: str,
        *,
        description: typing.Optional[str] = OMIT,
        name: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> Project:
        """
        Update a project's properties.

        Parameters
        ----------
        project_id : str

        description : typing.Optional[str]
            Optional project description.

        name : typing.Optional[str]
            Project name.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        Project
            Successful Response

        Examples
        --------
        import asyncio

        from terminaluse import AsyncTerminalUseBase

        client = AsyncTerminalUseBase(
            agent_api_key="YOUR_AGENT_API_KEY",
            token="YOUR_TOKEN",
            base_url="https://yourhost.com/path/to/api",
        )


        async def main() -> None:
            await client.projects.update(
                project_id="project_id",
            )


        asyncio.run(main())
        """
        _response = await self._raw_client.update(
            project_id, description=description, name=name, request_options=request_options
        )
        return _response.data

    async def list_collaborators(
        self, project_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> ProjectCollaboratorsResponse:
        """
        List explicit collaborator roles on a project.

        Parameters
        ----------
        project_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ProjectCollaboratorsResponse
            Successful Response

        Examples
        --------
        import asyncio

        from terminaluse import AsyncTerminalUseBase

        client = AsyncTerminalUseBase(
            agent_api_key="YOUR_AGENT_API_KEY",
            token="YOUR_TOKEN",
            base_url="https://yourhost.com/path/to/api",
        )


        async def main() -> None:
            await client.projects.list_collaborators(
                project_id="project_id",
            )


        asyncio.run(main())
        """
        _response = await self._raw_client.list_collaborators(project_id, request_options=request_options)
        return _response.data

    async def upsert_collaborator(
        self,
        project_id: str,
        member_id: str,
        *,
        role: UpsertProjectCollaboratorRequestRole,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> ProjectCollaboratorResponse:
        """
        Create or update an explicit collaborator role on a project.

        Parameters
        ----------
        project_id : str

        member_id : str

        role : UpsertProjectCollaboratorRequestRole
            Role to assign.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ProjectCollaboratorResponse
            Successful Response

        Examples
        --------
        import asyncio

        from terminaluse import AsyncTerminalUseBase

        client = AsyncTerminalUseBase(
            agent_api_key="YOUR_AGENT_API_KEY",
            token="YOUR_TOKEN",
            base_url="https://yourhost.com/path/to/api",
        )


        async def main() -> None:
            await client.projects.upsert_collaborator(
                project_id="project_id",
                member_id="member_id",
                role="discoverer",
            )


        asyncio.run(main())
        """
        _response = await self._raw_client.upsert_collaborator(
            project_id, member_id, role=role, request_options=request_options
        )
        return _response.data

    async def remove_collaborator(
        self, project_id: str, member_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> DeleteResponse:
        """
        Remove explicit collaborator role tuples from a project.

        Parameters
        ----------
        project_id : str

        member_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        DeleteResponse
            Successful Response

        Examples
        --------
        import asyncio

        from terminaluse import AsyncTerminalUseBase

        client = AsyncTerminalUseBase(
            agent_api_key="YOUR_AGENT_API_KEY",
            token="YOUR_TOKEN",
            base_url="https://yourhost.com/path/to/api",
        )


        async def main() -> None:
            await client.projects.remove_collaborator(
                project_id="project_id",
                member_id="member_id",
            )


        asyncio.run(main())
        """
        _response = await self._raw_client.remove_collaborator(project_id, member_id, request_options=request_options)
        return _response.data



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from pydantic import BaseModel

# Import StateMachine only for type checking to avoid circular imports
if TYPE_CHECKING:
    from terminaluse.lib.sdk.state_machine import StateMachine


class StateWorkflow(ABC):
    @abstractmethod
    async def execute(self, state_machine: "StateMachine", state_machine_data: BaseModel | None = None) -> str:
        pass



# This file was auto-generated by Fern from our API Definition.

import typing

from ..core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ..core.request_options import RequestOptions
from ..types.principal_context_entity import PrincipalContextEntity
from .raw_client import AsyncRawAuthenticationClient, RawAuthenticationClient


class AuthenticationClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._raw_client = RawAuthenticationClient(client_wrapper=client_wrapper)

    @property
    def with_raw_response(self) -> RawAuthenticationClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        RawAuthenticationClient
        """
        return self._raw_client

    def authenticate(self, *, request_options: typing.Optional[RequestOptions] = None) -> PrincipalContextEntity:
        """
        Validate a WorkOS JWT and return the principal context.

        Parameters
        ----------
        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        PrincipalContextEntity
            Successful Response

        Examples
        --------
        from terminaluse import TerminalUseBase

        client = TerminalUseBase(
            agent_api_key="YOUR_AGENT_API_KEY",
            token="YOUR_TOKEN",
            base_url="https://yourhost.com/path/to/api",
        )
        client.authentication.authenticate()
        """
        _response = self._raw_client.authenticate(request_options=request_options)
        return _response.data


class AsyncAuthenticationClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._raw_client = AsyncRawAuthenticationClient(client_wrapper=client_wrapper)

    @property
    def with_raw_response(self) -> AsyncRawAuthenticationClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        AsyncRawAuthenticationClient
        """
        return self._raw_client

    async def authenticate(self, *, request_options: typing.Optional[RequestOptions] = None) -> PrincipalContextEntity:
        """
        Validate a WorkOS JWT and return the principal context.

        Parameters
        ----------
        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        PrincipalContextEntity
            Successful Response

        Examples
        --------
        import asyncio

        from terminaluse import AsyncTerminalUseBase

        client = AsyncTerminalUseBase(
            agent_api_key="YOUR_AGENT_API_KEY",
            token="YOUR_TOKEN",
            base_url="https://yourhost.com/path/to/api",
        )


        async def main() -> None:
            await client.authentication.authenticate()


        asyncio.run(main())
        """
        _response = await self._raw_client.authenticate(request_options=request_options)
        return _response.data



# This file was auto-generated by Fern from our API Definition.

import typing
from json.decoder import JSONDecodeError

from ..core.api_error import ApiError
from ..core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ..core.http_response import AsyncHttpResponse, HttpResponse
from ..core.jsonable_encoder import jsonable_encoder
from ..core.pydantic_utilities import parse_obj_as
from ..core.request_options import RequestOptions
from ..errors.unprocessable_entity_error import UnprocessableEntityError
from ..types.access_group import AccessGroup
from ..types.access_group_list_response import AccessGroupListResponse
from ..types.access_group_manager import AccessGroupManager
from ..types.access_group_managers_response import AccessGroupManagersResponse
from ..types.access_group_member import AccessGroupMember
from ..types.access_group_members_response import AccessGroupMembersResponse
from ..types.access_group_visible_org import AccessGroupVisibleOrg
from ..types.access_group_visible_orgs_response import AccessGroupVisibleOrgsResponse
from ..types.delete_response import DeleteResponse
from ..types.http_validation_error import HttpValidationError

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class RawGroupsClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._client_wrapper = client_wrapper

    def list(
        self,
        *,
        owner_org_id: typing.Optional[str] = None,
        limit: typing.Optional[int] = None,
        page_number: typing.Optional[int] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[AccessGroupListResponse]:
        """
        List groups visible to the current principal.

        Parameters
        ----------
        owner_org_id : typing.Optional[str]
            Optional owner org filter.

        limit : typing.Optional[int]

        page_number : typing.Optional[int]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[AccessGroupListResponse]
            Successful Response
        """
        _response = self._client_wrapper.httpx_client.request(
            "groups",
            method="GET",
            params={
                "owner_org_id": owner_org_id,
                "limit": limit,
                "page_number": page_number,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    AccessGroupListResponse,
                    parse_obj_as(
                        type_=AccessGroupListResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        HttpValidationError,
                        parse_obj_as(
                            type_=HttpValidationError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def create(
        self,
        *,
        name: str,
        description: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[AccessGroup]:
        """
        Create an org-owned access group.

        Parameters
        ----------
        name : str
            Group name (unique within owner org).

        description : typing.Optional[str]
            Optional group description.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[AccessGroup]
            Successful Response
        """
        _response = self._client_wrapper.httpx_client.request(
            "groups",
            method="POST",
            json={
                "description": description,
                "name": name,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    AccessGroup,
                    parse_obj_as(
                        type_=AccessGroup,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        HttpValidationError,
                        parse_obj_as(
                            type_=HttpValidationError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def retrieve(
        self, group_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[AccessGroup]:
        """
        Get group details.

        Parameters
        ----------
        group_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[AccessGroup]
            Successful Response
        """
        _response = self._client_wrapper.httpx_client.request(
            f"groups/{jsonable_encoder(group_id)}",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    AccessGroup,
                    parse_obj_as(
                        type_=AccessGroup,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        HttpValidationError,
                        parse_obj_as(
                            type_=HttpValidationError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def delete(
        self, group_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[DeleteResponse]:
        """
        Delete an access group.

        Parameters
        ----------
        group_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[DeleteResponse]
            Successful Response
        """
        _response = self._client_wrapper.httpx_client.request(
            f"groups/{jsonable_encoder(group_id)}",
            method="DELETE",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    DeleteResponse,
                    parse_obj_as(
                        type_=DeleteResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        HttpValidationError,
                        parse_obj_as(
                            type_=HttpValidationError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def update(
        self,
        group_id: str,
        *,
        description: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[AccessGroup]:
        """
        Update group metadata.

        Parameters
        ----------
        group_id : str

        description : typing.Optional[str]
            Optional group description.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[AccessGroup]
            Successful Response
        """
        _response = self._client_wrapper.httpx_client.request(
            f"groups/{jsonable_encoder(group_id)}",
            method="PATCH",
            json={
                "description": description,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    AccessGroup,
                    parse_obj_as(
                        type_=AccessGroup,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        HttpValidationError,
                        parse_obj_as(
                            type_=HttpValidationError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def managers_list(
        self, group_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[AccessGroupManagersResponse]:
        """
        Parameters
        ----------
        group_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[AccessGroupManagersResponse]
            Successful Response
        """
        _response = self._client_wrapper.httpx_client.request(
            f"groups/{jsonable_encoder(group_id)}/managers",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    AccessGroupManagersResponse,
                    parse_obj_as(
                        type_=AccessGroupManagersResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        HttpValidationError,
                        parse_obj_as(
                            type_=HttpValidationError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def managers_upsert(
        self, group_id: str, member_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[AccessGroupManager]:
        """
        Parameters
        ----------
        group_id : str

        member_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[AccessGroupManager]
            Successful Response
        """
        _response = self._client_wrapper.httpx_client.request(
            f"groups/{jsonable_encoder(group_id)}/managers/{jsonable_encoder(member_id)}",
            method="PUT",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    AccessGroupManager,
                    parse_obj_as(
                        type_=AccessGroupManager,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        HttpValidationError,
                        parse_obj_as(
                            type_=HttpValidationError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def managers_delete(
        self, group_id: str, member_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[DeleteResponse]:
        """
        Parameters
        ----------
        group_id : str

        member_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[DeleteResponse]
            Successful Response
        """
        _response = self._client_wrapper.httpx_client.request(
            f"groups/{jsonable_encoder(group_id)}/managers/{jsonable_encoder(member_id)}",
            method="DELETE",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    DeleteResponse,
                    parse_obj_as(
                        type_=DeleteResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        HttpValidationError,
                        parse_obj_as(
                            type_=HttpValidationError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def members_list(
        self, group_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[AccessGroupMembersResponse]:
        """
        Parameters
        ----------
        group_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[AccessGroupMembersResponse]
            Successful Response
        """
        _response = self._client_wrapper.httpx_client.request(
            f"groups/{jsonable_encoder(group_id)}/members",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    AccessGroupMembersResponse,
                    parse_obj_as(
                        type_=AccessGroupMembersResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        HttpValidationError,
                        parse_obj_as(
                            type_=HttpValidationError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def members_upsert(
        self, group_id: str, member_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[AccessGroupMember]:
        """
        Parameters
        ----------
        group_id : str

        member_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[AccessGroupMember]
            Successful Response
        """
        _response = self._client_wrapper.httpx_client.request(
            f"groups/{jsonable_encoder(group_id)}/members/{jsonable_encoder(member_id)}",
            method="PUT",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    AccessGroupMember,
                    parse_obj_as(
                        type_=AccessGroupMember,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        HttpValidationError,
                        parse_obj_as(
                            type_=HttpValidationError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def members_delete(
        self, group_id: str, member_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[DeleteResponse]:
        """
        Parameters
        ----------
        group_id : str

        member_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[DeleteResponse]
            Successful Response
        """
        _response = self._client_wrapper.httpx_client.request(
            f"groups/{jsonable_encoder(group_id)}/members/{jsonable_encoder(member_id)}",
            method="DELETE",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    DeleteResponse,
                    parse_obj_as(
                        type_=DeleteResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        HttpValidationError,
                        parse_obj_as(
                            type_=HttpValidationError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def visible_orgs_list(
        self, group_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[AccessGroupVisibleOrgsResponse]:
        """
        Parameters
        ----------
        group_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[AccessGroupVisibleOrgsResponse]
            Successful Response
        """
        _response = self._client_wrapper.httpx_client.request(
            f"groups/{jsonable_encoder(group_id)}/visible-orgs",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    AccessGroupVisibleOrgsResponse,
                    parse_obj_as(
                        type_=AccessGroupVisibleOrgsResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        HttpValidationError,
                        parse_obj_as(
                            type_=HttpValidationError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def visible_orgs_upsert(
        self, group_id: str, org_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[AccessGroupVisibleOrg]:
        """
        Parameters
        ----------
        group_id : str

        org_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[AccessGroupVisibleOrg]
            Successful Response
        """
        _response = self._client_wrapper.httpx_client.request(
            f"groups/{jsonable_encoder(group_id)}/visible-orgs/{jsonable_encoder(org_id)}",
            method="PUT",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    AccessGroupVisibleOrg,
                    parse_obj_as(
                        type_=AccessGroupVisibleOrg,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        HttpValidationError,
                        parse_obj_as(
                            type_=HttpValidationError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def visible_orgs_delete(
        self, group_id: str, org_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[DeleteResponse]:
        """
        Parameters
        ----------
        group_id : str

        org_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[DeleteResponse]
            Successful Response
        """
        _response = self._client_wrapper.httpx_client.request(
            f"groups/{jsonable_encoder(group_id)}/visible-orgs/{jsonable_encoder(org_id)}",
            method="DELETE",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    DeleteResponse,
                    parse_obj_as(
                        type_=DeleteResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        HttpValidationError,
                        parse_obj_as(
                            type_=HttpValidationError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)


class AsyncRawGroupsClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._client_wrapper = client_wrapper

    async def list(
        self,
        *,
        owner_org_id: typing.Optional[str] = None,
        limit: typing.Optional[int] = None,
        page_number: typing.Optional[int] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[AccessGroupListResponse]:
        """
        List groups visible to the current principal.

        Parameters
        ----------
        owner_org_id : typing.Optional[str]
            Optional owner org filter.

        limit : typing.Optional[int]

        page_number : typing.Optional[int]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[AccessGroupListResponse]
            Successful Response
        """
        _response = await self._client_wrapper.httpx_client.request(
            "groups",
            method="GET",
            params={
                "owner_org_id": owner_org_id,
                "limit": limit,
                "page_number": page_number,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    AccessGroupListResponse,
                    parse_obj_as(
                        type_=AccessGroupListResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        HttpValidationError,
                        parse_obj_as(
                            type_=HttpValidationError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def create(
        self,
        *,
        name: str,
        description: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[AccessGroup]:
        """
        Create an org-owned access group.

        Parameters
        ----------
        name : str
            Group name (unique within owner org).

        description : typing.Optional[str]
            Optional group description.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[AccessGroup]
            Successful Response
        """
        _response = await self._client_wrapper.httpx_client.request(
            "groups",
            method="POST",
            json={
                "description": description,
                "name": name,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    AccessGroup,
                    parse_obj_as(
                        type_=AccessGroup,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        HttpValidationError,
                        parse_obj_as(
                            type_=HttpValidationError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def retrieve(
        self, group_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[AccessGroup]:
        """
        Get group details.

        Parameters
        ----------
        group_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[AccessGroup]
            Successful Response
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"groups/{jsonable_encoder(group_id)}",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    AccessGroup,
                    parse_obj_as(
                        type_=AccessGroup,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        HttpValidationError,
                        parse_obj_as(
                            type_=HttpValidationError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def delete(
        self, group_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[DeleteResponse]:
        """
        Delete an access group.

        Parameters
        ----------
        group_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[DeleteResponse]
            Successful Response
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"groups/{jsonable_encoder(group_id)}",
            method="DELETE",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    DeleteResponse,
                    parse_obj_as(
                        type_=DeleteResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        HttpValidationError,
                        parse_obj_as(
                            type_=HttpValidationError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def update(
        self,
        group_id: str,
        *,
        description: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[AccessGroup]:
        """
        Update group metadata.

        Parameters
        ----------
        group_id : str

        description : typing.Optional[str]
            Optional group description.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[AccessGroup]
            Successful Response
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"groups/{jsonable_encoder(group_id)}",
            method="PATCH",
            json={
                "description": description,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    AccessGroup,
                    parse_obj_as(
                        type_=AccessGroup,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        HttpValidationError,
                        parse_obj_as(
                            type_=HttpValidationError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def managers_list(
        self, group_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[AccessGroupManagersResponse]:
        """
        Parameters
        ----------
        group_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[AccessGroupManagersResponse]
            Successful Response
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"groups/{jsonable_encoder(group_id)}/managers",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    AccessGroupManagersResponse,
                    parse_obj_as(
                        type_=AccessGroupManagersResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        HttpValidationError,
                        parse_obj_as(
                            type_=HttpValidationError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def managers_upsert(
        self, group_id: str, member_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[AccessGroupManager]:
        """
        Parameters
        ----------
        group_id : str

        member_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[AccessGroupManager]
            Successful Response
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"groups/{jsonable_encoder(group_id)}/managers/{jsonable_encoder(member_id)}",
            method="PUT",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    AccessGroupManager,
                    parse_obj_as(
                        type_=AccessGroupManager,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        HttpValidationError,
                        parse_obj_as(
                            type_=HttpValidationError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def managers_delete(
        self, group_id: str, member_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[DeleteResponse]:
        """
        Parameters
        ----------
        group_id : str

        member_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[DeleteResponse]
            Successful Response
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"groups/{jsonable_encoder(group_id)}/managers/{jsonable_encoder(member_id)}",
            method="DELETE",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    DeleteResponse,
                    parse_obj_as(
                        type_=DeleteResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        HttpValidationError,
                        parse_obj_as(
                            type_=HttpValidationError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def members_list(
        self, group_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[AccessGroupMembersResponse]:
        """
        Parameters
        ----------
        group_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[AccessGroupMembersResponse]
            Successful Response
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"groups/{jsonable_encoder(group_id)}/members",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    AccessGroupMembersResponse,
                    parse_obj_as(
                        type_=AccessGroupMembersResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        HttpValidationError,
                        parse_obj_as(
                            type_=HttpValidationError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def members_upsert(
        self, group_id: str, member_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[AccessGroupMember]:
        """
        Parameters
        ----------
        group_id : str

        member_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[AccessGroupMember]
            Successful Response
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"groups/{jsonable_encoder(group_id)}/members/{jsonable_encoder(member_id)}",
            method="PUT",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    AccessGroupMember,
                    parse_obj_as(
                        type_=AccessGroupMember,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        HttpValidationError,
                        parse_obj_as(
                            type_=HttpValidationError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def members_delete(
        self, group_id: str, member_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[DeleteResponse]:
        """
        Parameters
        ----------
        group_id : str

        member_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[DeleteResponse]
            Successful Response
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"groups/{jsonable_encoder(group_id)}/members/{jsonable_encoder(member_id)}",
            method="DELETE",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    DeleteResponse,
                    parse_obj_as(
                        type_=DeleteResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        HttpValidationError,
                        parse_obj_as(
                            type_=HttpValidationError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def visible_orgs_list(
        self, group_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[AccessGroupVisibleOrgsResponse]:
        """
        Parameters
        ----------
        group_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[AccessGroupVisibleOrgsResponse]
            Successful Response
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"groups/{jsonable_encoder(group_id)}/visible-orgs",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    AccessGroupVisibleOrgsResponse,
                    parse_obj_as(
                        type_=AccessGroupVisibleOrgsResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        HttpValidationError,
                        parse_obj_as(
                            type_=HttpValidationError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def visible_orgs_upsert(
        self, group_id: str, org_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[AccessGroupVisibleOrg]:
        """
        Parameters
        ----------
        group_id : str

        org_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[AccessGroupVisibleOrg]
            Successful Response
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"groups/{jsonable_encoder(group_id)}/visible-orgs/{jsonable_encoder(org_id)}",
            method="PUT",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    AccessGroupVisibleOrg,
                    parse_obj_as(
                        type_=AccessGroupVisibleOrg,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        HttpValidationError,
                        parse_obj_as(
                            type_=HttpValidationError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def visible_orgs_delete(
        self, group_id: str, org_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[DeleteResponse]:
        """
        Parameters
        ----------
        group_id : str

        org_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[DeleteResponse]
            Successful Response
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"groups/{jsonable_encoder(group_id)}/visible-orgs/{jsonable_encoder(org_id)}",
            method="DELETE",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    DeleteResponse,
                    parse_obj_as(
                        type_=DeleteResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        HttpValidationError,
                        parse_obj_as(
                            type_=HttpValidationError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)



# This file was auto-generated by Fern from our API Definition.

import datetime as dt
import typing

import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel


class FilesystemFileResponse(UniversalBaseModel):
    """
    File metadata response - adds server-side fields.
    """

    checksum: typing.Optional[str] = pydantic.Field(default=None)
    """
    SHA256 checksum (None for directories).
    """

    content: typing.Optional[str] = pydantic.Field(default=None)
    """
    File contents (None for binary/large files).
    """

    content_truncated: typing.Optional[bool] = pydantic.Field(default=None)
    """
    True if content was truncated due to size.
    """

    is_binary: typing.Optional[bool] = pydantic.Field(default=None)
    """
    True if file is binary.
    """

    is_directory: typing.Optional[bool] = pydantic.Field(default=None)
    """
    Whether this entry is a directory.
    """

    mime_type: typing.Optional[str] = pydantic.Field(default=None)
    """
    MIME type (None for directories).
    """

    modified_at: dt.datetime = pydantic.Field()
    """
    File's last modification time.
    """

    path: str = pydantic.Field()
    """
    Relative path within filesystem (e.g., 'src/index.ts').
    """

    size_bytes: typing.Optional[int] = pydantic.Field(default=None)
    """
    File size in bytes (None for directories).
    """

    synced_at: dt.datetime = pydantic.Field()
    """
    When this entry was last synced.
    """

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



"""Tests for CLI version check utilities."""

from __future__ import annotations

import json
import time
from unittest.mock import patch


def test_parse_version_simple():
    from terminaluse.lib.cli.utils.version_check import _parse_version

    assert _parse_version("0.25.0") == (0, 25, 0)
    assert _parse_version("1.0.0") == (1, 0, 0)
    assert _parse_version("10.20.30") == (10, 20, 30)


def test_parse_version_prerelease():
    from terminaluse.lib.cli.utils.version_check import _parse_version

    assert _parse_version("0.25.0a1") == (0, 25, 0)
    assert _parse_version("1.0.0b2") == (1, 0, 0)
    assert _parse_version("2.0.0rc1") == (2, 0, 0)


def test_parse_version_invalid():
    from terminaluse.lib.cli.utils.version_check import _parse_version

    assert _parse_version("invalid") == (0, 0, 0)
    assert _parse_version("") == (0, 0, 0)


def test_is_newer():
    from terminaluse.lib.cli.utils.version_check import _is_newer

    assert _is_newer("0.26.0", "0.25.0") is True
    assert _is_newer("1.0.0", "0.99.99") is True
    assert _is_newer("0.25.1", "0.25.0") is True

    assert _is_newer("0.25.0", "0.25.0") is False
    assert _is_newer("0.24.0", "0.25.0") is False
    assert _is_newer("0.25.0", "1.0.0") is False


def test_check_for_updates_uses_cache(tmp_path, monkeypatch):
    """When cache is fresh, should return cached version without network call."""
    from terminaluse.lib.cli.utils import version_check

    cache_file = tmp_path / "version_cache.json"
    monkeypatch.setattr(version_check, "VERSION_CACHE_FILE", cache_file)
    monkeypatch.setattr(version_check, "__version__", "0.25.0")

    # Write fresh cache with newer version
    cache_file.write_text(
        json.dumps(
            {
                "latest_version": "0.26.0",
                "checked_at": time.time(),
            }
        )
    )

    with patch.object(version_check, "_fetch_latest_version") as mock_fetch:
        result = version_check.check_for_updates()

        assert result.available == "0.26.0"
        assert result.checked is True
        mock_fetch.assert_not_called()


def test_check_for_updates_stale_cache_fetches(tmp_path, monkeypatch):
    """When cache is stale, should fetch from PyPI."""
    from terminaluse.lib.cli.utils import version_check

    cache_file = tmp_path / "version_cache.json"
    monkeypatch.setattr(version_check, "VERSION_CACHE_FILE", cache_file)
    monkeypatch.setattr(version_check, "__version__", "0.25.0")

    # Write stale cache (2 hours old)
    cache_file.write_text(
        json.dumps(
            {
                "latest_version": "0.25.0",
                "checked_at": time.time() - 2 * 60 * 60,
            }
        )
    )

    with patch.object(version_check, "_fetch_latest_version", return_value="0.27.0") as mock_fetch:
        result = version_check.check_for_updates()

        assert result.available == "0.27.0"
        assert result.checked is True
        mock_fetch.assert_called_once()


def test_check_for_updates_no_update_available(tmp_path, monkeypatch):
    """When current version is latest, should return no available update."""
    from terminaluse.lib.cli.utils import version_check

    cache_file = tmp_path / "version_cache.json"
    monkeypatch.setattr(version_check, "VERSION_CACHE_FILE", cache_file)
    monkeypatch.setattr(version_check, "__version__", "0.25.0")

    cache_file.write_text(
        json.dumps(
            {
                "latest_version": "0.25.0",
                "checked_at": time.time(),
            }
        )
    )

    result = version_check.check_for_updates()
    assert result.available is None
    assert result.checked is True


def test_check_for_updates_network_failure(tmp_path, monkeypatch):
    """Network failures should set checked=False."""
    from terminaluse.lib.cli.utils import version_check

    cache_file = tmp_path / "version_cache.json"
    monkeypatch.setattr(version_check, "VERSION_CACHE_FILE", cache_file)

    with patch.object(version_check, "_fetch_latest_version", return_value=None):
        result = version_check.check_for_updates()
        assert result.available is None
        assert result.checked is False


def test_check_for_updates_force_bypasses_cache(tmp_path, monkeypatch):
    """When force=True, should fetch from PyPI even if cache is fresh."""
    from terminaluse.lib.cli.utils import version_check

    cache_file = tmp_path / "version_cache.json"
    monkeypatch.setattr(version_check, "VERSION_CACHE_FILE", cache_file)
    monkeypatch.setattr(version_check, "__version__", "0.25.0")

    # Write fresh cache
    cache_file.write_text(
        json.dumps(
            {
                "latest_version": "0.25.0",
                "checked_at": time.time(),
            }
        )
    )

    with patch.object(version_check, "_fetch_latest_version", return_value="0.28.0") as mock_fetch:
        result = version_check.check_for_updates(force=True)

        assert result.available == "0.28.0"
        assert result.checked is True
        mock_fetch.assert_called_once()
